import matplotlib
import numpy as np

import matplotlib.pyplot as plt
from matplotlib import cm

from rlberry.envs.finite_mdp import FiniteMDP
from rlberry_scool.envs.finite import gridworld_utils
from rlberry.rendering import Scene, GeometricPrimitive, RenderInterface2D
from rlberry.rendering.common_shapes import circle_shape


import rlberry

logger = rlberry.logger


class GridWorld(RenderInterface2D, FiniteMDP):
    """
    Simple GridWorld environment.

    Parameters
    -----------
    nrows : int
        number of rows
    ncols : int
        number of columns
    start_coord : tuple
        tuple with coordinates of initial position
    terminal_states : tuple
        ((row_0, col_0), (row_1, col_1), ...) = coordinates of
        terminal states
    success_probability : double
        probability of moving in the chosen direction
    reward_at: dict
        dictionary, keys = tuple containing coordinates, values = reward
        at each coordinate
    walls : tuple
        ((row_0, col_0), (row_1, col_1), ...) = coordinates of walls
    default_reward : double
        reward received at states not in  'reward_at'

    """

    name = "GridWorld"

    def __init__(
        self,
        nrows=5,
        ncols=5,
        start_coord=(0, 0),
        terminal_states=None,
        success_probability=0.9,
        reward_at=None,
        walls=((1, 1), (2, 2)),
        default_reward=0.0,
    ):
        # Grid dimensions
        self.nrows = nrows
        self.ncols = ncols

        # Reward parameters
        self.default_reward = default_reward

        # Default config
        if reward_at is not None:
            self.reward_at = reward_at
        else:
            self.reward_at = {(nrows - 1, ncols - 1): 1}
        if walls is not None:
            self.walls = walls
        else:
            self.walls = ()
        if terminal_states is not None:
            self.terminal_states = terminal_states
        else:
            self.terminal_states = ()

        # Probability of going left/right/up/down when choosing the
        # correspondent action
        # The remaining probability mass is distributed uniformly to other
        # available actions
        self.success_probability = success_probability

        # Start coordinate
        self.start_coord = tuple(start_coord)

        # Actions (string to index & index to string)
        self.a_str2idx = {"left": 0, "right": 1, "down": 2, "up": 3}
        self.a_idx2str = {0: "left", 1: "right", 2: "down", 3: "up"}

        # --------------------------------------------
        # The variables below are defined in _build()
        # --------------------------------------------

        # Mappings (state index) <-> (state coordinate)
        self.index2coord = {}
        self.coord2index = {}

        # MDP parameters for base class
        self.P = None
        self.R = None
        self.Ns = None
        self.Na = 4

        # Build
        self._build()
        init_state_idx = self.coord2index[start_coord]
        FiniteMDP.__init__(
            self, self.R, self.P, initial_state_distribution=init_state_idx
        )
        RenderInterface2D.__init__(self)
        self.reset()
        self.reward_range = (self.R.min(), self.R.max())

        # rendering info
        self.set_clipping_area((0, self.ncols, 0, self.nrows))
        self.set_refresh_interval(100)  # in milliseconds
        self.renderer_type = "opengl"

    @classmethod
    def from_layout(
        cls, layout: str = gridworld_utils.DEFAULT_LAYOUT, success_probability=0.95
    ):
        """
        Create GridWorld instance from a layout.

        Layout symbols:

        '#' : wall
        'r' : reward of 1, terminal state
        'R' : reward of 1, non-terminal state
        'T' : terminal state
        'I' : initial state (if several, start uniformly among I)
        'O' : empty state
        any other char : empty state

        Layout example:

        IOOOO # OOOOO  O OOOOR
        OOOOO # OOOOO  # OOOOO
        OOOOO O OOOOO  # OOOOO
        OOOOO # OOOOO  # OOOOO
        IOOOO # OOOOO  # OOOOr
        """
        info = gridworld_utils.get_layout_info(layout)
        nrows = info["nrows"]
        ncols = info["ncols"]
        walls = info["walls"]
        reward_at = info["reward_at"]
        terminal_states = info["terminal_states"]
        initial_states_coord = info["initial_states"]

        # Init base class
        env = cls(
            nrows=nrows,
            ncols=ncols,
            terminal_states=terminal_states,
            success_probability=success_probability,
            reward_at=reward_at,
            walls=walls,
            default_reward=0.0,
        )

        # Set initial distribution
        distr = np.zeros(env.observation_space.n)
        for init_coord in initial_states_coord:
            init_index = env.coord2index[init_coord]
            distr[init_index] = 1.0
        distr = distr / distr.sum()
        env.set_initial_state_distribution(distr)

        return env

    def is_terminal(self, state):
        state_coord = self.index2coord[state]
        return state_coord in self.terminal_states

    def reward_fn(self, state, action, next_state):
        row, col = self.index2coord[state]
        if (row, col) in self.reward_at:
            return self.reward_at[(row, col)]
        if (row, col) in self.walls:
            return 0.0
        return self.default_reward

    def _build(self):
        self._build_state_mappings_and_states()
        self._build_transition_probabilities()
        self._build_mean_rewards()

    def _build_state_mappings_and_states(self):
        index = 0
        for rr in range(self.nrows):
            for cc in range(self.ncols):
                if (rr, cc) in self.walls:
                    self.coord2index[(rr, cc)] = -1
                else:
                    self.coord2index[(rr, cc)] = index
                    self.index2coord[index] = (rr, cc)
                    index += 1
        states = np.arange(index).tolist()
        self.Ns = len(states)

    def _build_mean_rewards(self):
        S = self.Ns
        A = self.Na
        self.R = np.zeros((S, A))
        for ss in range(S):
            for aa in range(A):
                mean_r = 0
                for ns in range(S):
                    mean_r += self.reward_fn(ss, aa, ns) * self.P[ss, aa, ns]
                self.R[ss, aa] = mean_r

    def _build_transition_probabilities(self):
        Ns = self.Ns
        Na = self.Na
        self.P = np.zeros((Ns, Na, Ns))
        for s in range(Ns):
            s_coord = self.index2coord[s]
            neighbors = self._get_neighbors(*s_coord)
            valid_neighbors = [neighbors[nn][0] for nn in neighbors if neighbors[nn][1]]
            n_valid = len(valid_neighbors)
            for a in range(Na):  # each action corresponds to a direction
                for nn in neighbors:
                    next_s_coord = neighbors[nn][0]
                    if next_s_coord in valid_neighbors:
                        next_s = self.coord2index[next_s_coord]
                        if a == nn:  # action is successful
                            self.P[s, a, next_s] = self.success_probability + (
                                1 - self.success_probability
                            ) * (n_valid == 1)
                        elif neighbors[a][0] not in valid_neighbors:
                            self.P[s, a, s] = 1.0
                        else:
                            if n_valid > 1:
                                self.P[s, a, next_s] = (
                                    1.0 - self.success_probability
                                ) / (n_valid - 1)

    def _get_neighbors(self, row, col):
        aux = {}
        aux["left"] = (row, col - 1)  # left
        aux["right"] = (row, col + 1)  # right
        aux["up"] = (row - 1, col)  # up
        aux["down"] = (row + 1, col)  # down
        neighbors = {}
        for direction_str in aux:
            direction = self.a_str2idx[direction_str]
            next_s = aux[direction_str]
            neighbors[direction] = (next_s, self._is_valid(*next_s))
        return neighbors

    def get_transition_support(self, state):
        row, col = self.index2coord[state]
        neighbors = [(row, col - 1), (row, col + 1), (row - 1, col), (row + 1, col)]
        return [
            self.coord2index[coord] for coord in neighbors if self._is_valid(*coord)
        ]

    def _is_valid(self, row, col):
        if (row, col) in self.walls:
            return False
        elif row < 0 or row >= self.nrows:
            return False
        elif col < 0 or col >= self.ncols:
            return False
        return True

    def _build_ascii(self):
        grid = [[""] * self.ncols for rr in range(self.nrows)]
        grid_idx = [[""] * self.ncols for rr in range(self.nrows)]
        for rr in range(self.nrows):
            for cc in range(self.ncols):
                if (rr, cc) in self.walls:
                    grid[rr][cc] = "x "
                else:
                    grid[rr][cc] = "o "
                grid_idx[rr][cc] = str(self.coord2index[(rr, cc)]).zfill(3)

        for rr, cc in self.reward_at:
            rwd = self.reward_at[(rr, cc)]
            if rwd > 0:
                grid[rr][cc] = "+ "
            if rwd < 0:
                grid[rr][cc] = "-"

        grid[self.start_coord[0]][self.start_coord[1]] = "I "

        # current position of the agent
        x, y = self.index2coord[self.state]
        grid[x][y] = "A "

        #
        grid_ascii = ""
        for rr in range(self.nrows + 1):
            if rr < self.nrows:
                grid_ascii += str(rr).zfill(2) + 2 * " " + " ".join(grid[rr]) + "\n"
            else:
                grid_ascii += 3 * " " + " ".join(
                    [str(jj).zfill(2) for jj in range(self.ncols)]
                )

        self.grid_ascii = grid_ascii
        self.grid_idx = grid_idx
        return self.grid_ascii

    def display_values(self, values):
        assert len(values) == self.Ns
        grid_values = [["X".ljust(9)] * self.ncols for ii in range(self.nrows)]
        for s_idx in range(self.Ns):
            v = values[s_idx]
            row, col = self.index2coord[s_idx]
            grid_values[row][col] = ("%0.2f" % v).ljust(9)

        grid_values_ascii = ""
        for rr in range(self.nrows + 1):
            if rr < self.nrows:
                grid_values_ascii += (
                    str(rr).zfill(2) + 2 * " " + " ".join(grid_values[rr]) + "\n"
                )
            else:
                grid_values_ascii += 4 * " " + " ".join(
                    [str(jj).zfill(2).ljust(9) for jj in range(self.ncols)]
                )
        logger.info(grid_values_ascii)

    def print_transition_at(self, row, col, action):
        s_idx = self.coord2index[(row, col)]
        if s_idx < 0:
            logger.info("wall!")
            return
        a_idx = self.a_str2idx[action]
        for next_s_idx, prob in enumerate(self.P[s_idx, a_idx]):
            if prob > 0:
                logger.info(
                    "to (%d, %d) with prob %f"
                    % (self.index2coord[next_s_idx] + (prob,))
                )

    def render_ascii(self):
        print(self._build_ascii())

    def step(self, action):
        assert self.action_space.contains(action), "Invalid action!"

        # save state for rendering
        if self.is_render_enabled():
            self.append_state_for_rendering(self.state)

        # take step
        next_state, reward, terminated, truncated, info = self.sample(
            self.state, action
        )
        self.state = next_state
        return next_state, reward, terminated, truncated, info

    #
    # Code for rendering
    #
    def get_layout_array(self, state_data=None, fill_walls_with=np.nan):
        """
        Returns an array 'layout' of shape (nrows, ncols) such that:

            layout[row, col] = state_data[self.coord2idx[row, col]]

        If (row, col) is a wall:

            layout[row, col] = fill_walls_with

        Parameters
        ----------
        state_data : np.array, default = None
            Array of shape (self.observation_space.n,)
        fill_walls_with : float, default: np.nan
            Value to set in the layout in the coordinates corresponding to walls.

        Returns
        -------
        Gridworld layout array of shape (nrows, ncols).
        """
        layout = np.zeros((self.nrows, self.ncols))
        if state_data is not None:
            assert state_data.shape == (self.observation_space.n,)
            data_rows = [self.index2coord[idx][0] for idx in self.index2coord]
            data_cols = [self.index2coord[idx][1] for idx in self.index2coord]
            layout[data_rows, data_cols] = state_data
        else:
            state_rr, state_cc = self.index2coord[self.state]
            layout[state_rr, state_cc] = 1.0

        walls_rows = [ww[0] for ww in self.walls]
        walls_cols = [ww[1] for ww in self.walls]
        layout[walls_rows, walls_cols] = fill_walls_with
        return layout

    def get_layout_img(
        self, state_data=None, colormap_name="cool", wall_color=(0.0, 0.0, 0.0)
    ):
        """
        Returns an image array representing the value of `state_data` on
        the gridworld layout.

        Parameters
        ----------
        state_data : np.array, default = None
            Array of shape (self.observation_space.n,)
        colormap_name : str, default = 'cool'
            Colormap name.
            See https://matplotlib.org/tutorials/colors/colormaps.html
        wall_color : tuple
            RGB color for walls.
        Returns
        -------
        Gridworld image array of shape (nrows, ncols, 3).
        """
        # map data to [0.0, 1.0]
        if state_data is not None:
            state_data = state_data - state_data.min()
            if state_data.max() > 0.0:
                state_data = state_data / state_data.max()

        colormap_fn = plt.get_cmap(colormap_name)
        layout = self.get_layout_array(state_data, fill_walls_with=np.nan)
        norm = matplotlib.colors.Normalize(vmin=0.0, vmax=1.0)
        scalar_map = cm.ScalarMappable(norm=norm, cmap=colormap_fn)
        img = np.zeros(layout.shape + (3,))
        for rr in range(layout.shape[0]):
            for cc in range(layout.shape[1]):
                if np.isnan(layout[rr, cc]):
                    img[self.nrows - 1 - rr, cc, :] = wall_color
                else:
                    img[self.nrows - 1 - rr, cc, :3] = scalar_map.to_rgba(
                        layout[rr, cc]
                    )[:3]
        return img

    def get_background(self):
        """
        Return a scene (list of shapes) representing the background
        """
        bg = Scene()

        # walls
        for wall in self.walls:
            y, x = wall
            shape = GeometricPrimitive("POLYGON")
            shape.set_color((0.25, 0.25, 0.25))
            shape.add_vertex((x, y))
            shape.add_vertex((x + 1, y))
            shape.add_vertex((x + 1, y + 1))
            shape.add_vertex((x, y + 1))
            bg.add_shape(shape)

        # rewards
        for y, x in self.reward_at:
            flag = GeometricPrimitive("POLYGON")
            rwd = self.reward_at[(y, x)]
            color = 0.5 * np.abs(rwd) / self.reward_range[1]
            if rwd > 0:
                flag.set_color((0.0, color, 0.0))
            if rwd < 0:
                flag.set_color((color, 0.0, 0.0))

            x += 0.5
            y += 0.25
            flag.add_vertex((x, y))
            flag.add_vertex((x + 0.25, y + 0.5))
            flag.add_vertex((x - 0.25, y + 0.5))
            bg.add_shape(flag)

        return bg

    def get_scene(self, state):
        """
        Return scene (list of shapes) representing a given state
        """
        y, x = self.index2coord[state]
        x = x + 0.5  # centering
        y = y + 0.5  # centering

        scene = Scene()

        agent = circle_shape((x, y), 0.25, n_points=5)
        agent.type = "POLYGON"
        agent.set_color((0.75, 0.0, 0.5))

        scene.add_shape(agent)
        return scene



import numpy as np

WALL_SYMBOL = "#"
REWARD_TERMINAL_SYMBOL = "r"
REWARD_SYMBOL = "R"
TERMINAL_STATE_SYMBOL = "T"
INITIAL_STATE_SYMBOL = "I"


# spaces are ignored
DEFAULT_LAYOUT = """
IOOOO # OOOOO  O OOOOR
OOOOO # OOOOO  # OOOOO
OOOOO O OOOOO  # OOOOO
OOOOO # OOOOO  # OOOOO
IOOOO # OOOOO  # OOOOr
"""


def _preprocess_layout(layout):
    layout = layout.replace(" ", "")  # remove spaces
    # remove first and last line breaks
    if layout[0] == "\n":
        layout = layout[1:]
    if layout[-1] == "\n":
        layout = layout[:-1]

    # make sure all lines have the same length
    lines = layout.split("\n")
    len_lines = [len(line) for line in lines]
    max_len = np.max(len_lines)
    # below, also reverse lines (so that render is not inversed in the y-direction)
    adjusted_lines = [
        line.ljust(max_len, "O") for line in reversed(lines)
    ]  # fill with empty state
    layout = "\n".join(adjusted_lines)
    return layout


def get_layout_info(layout):
    layout = _preprocess_layout(layout)
    lines = layout.split("\n")
    nrows = len(lines)
    ncols = len(lines[0])
    walls = []
    initial_states = []
    terminal_states = []
    reward_at = dict()
    for rr in range(nrows):
        line = lines[rr]
        for cc in range(ncols):
            symbol = line[cc]
            state_coord = (rr, cc)
            if symbol == WALL_SYMBOL:
                walls.append(state_coord)
            if symbol == TERMINAL_STATE_SYMBOL or symbol == REWARD_TERMINAL_SYMBOL:
                terminal_states.append(state_coord)
            if symbol == REWARD_SYMBOL or symbol == REWARD_TERMINAL_SYMBOL:
                reward_at[state_coord] = 1.0
            if symbol == INITIAL_STATE_SYMBOL:
                initial_states.append(state_coord)
    info = dict(
        nrows=nrows,
        ncols=ncols,
        initial_states=tuple(initial_states),
        terminal_states=tuple(terminal_states),
        walls=tuple(walls),
        reward_at=reward_at,
    )
    return info



import numpy as np


from rlberry.agents import AgentWithSimplePolicy
from rlberry_scool.agents.dynprog.utils import backward_induction, value_iteration
from gymnasium.spaces import Discrete

import rlberry

logger = rlberry.logger


class MBQVIAgent(AgentWithSimplePolicy):
    """
    Model-Basel Q-Value iteration (MBQVI).

    Builds an empirical MDP and runs value iteration on it.
    Corresponds to the "indirect" algorithm studied by Kearns and Singh (1999).

    Parameters
    -----------
    env : Model
        generative model with finite state-action space
    n_samples : int
        number of samples *per state-action pair* used to estimate
        the empirical MDP.
    gamma : double
        discount factor in [0, 1]
    horizon : int
        horizon, if the problem is finite-horizon. if None, the discounted
        problem is solved. default = None
    epsilon : double
        precision of value iteration, only used in discounted problems
        (when horizon is None).
    **kwargs : Keyword Arguments
         Arguments to be passed to `AgentWithSimplePolicy.__init__(self, env, **kwargs)` (:class:`~rlberry.agents.AgentWithSimplePolicy`).

    References
    ----------
    Kearns, Michael J., and Satinder P. Singh.
    "Finite-sample convergence rates for Q-learning and indirect algorithms."
    Advances in neural information processing systems. 1999.
    """

    name = "MBQVI"

    def __init__(
        self, env, n_samples=10, gamma=0.99, horizon=None, epsilon=1e-6, **kwargs
    ):
        AgentWithSimplePolicy.__init__(self, env, **kwargs)

        # initialize base class
        assert self.env.is_generative(), "MBQVI requires a generative model."
        assert isinstance(
            self.env.observation_space, Discrete
        ), "MBQVI requires a finite state space."
        assert isinstance(
            self.env.action_space, Discrete
        ), "MBQVI requires a finite action space."

        #
        self.n_samples = n_samples
        self.gamma = gamma
        self.horizon = horizon
        self.epsilon = epsilon

        # empirical MDP, created in fit()
        self.R_hat = None
        self.P_hat = None

        # value functions
        self.V = None
        self.Q = None

    def _update(self, state, action, next_state, reward):
        """Update model statistics."""
        self.N_sa[state, action] += 1
        self.N_sas[state, action, next_state] += 1
        self.S_sa[state, action] += reward

    def fit(self, budget=None, **kwargs):
        """
        Build empirical MDP and run value iteration.

        Parameters
        ----------
        budget: None
            Not used. Only defined for compatibility purpose with rlberry.
            Changing `budget` value has no effect.
        **kwargs : Keyword Arguments
            Extra arguments. Not used for this agent.
        """
        del kwargs
        S = self.env.observation_space.n
        A = self.env.action_space.n
        self.N_sa = np.zeros((S, A))
        self.N_sas = np.zeros((S, A, S))
        self.S_sa = np.zeros((S, A))

        # collect data
        total_samples = S * A * self.n_samples
        count = 0
        logger.debug(
            f"[{self.name}] collecting {self.n_samples} samples per (s,a)"
            f", total = {total_samples} samples."
        )
        for ss in range(S):
            for aa in range(A):
                for _ in range(self.n_samples):
                    next_state, reward, _, _, _ = self.env.sample(ss, aa)
                    self._update(ss, aa, next_state, reward)

                    count += 1
                    if count % 10000 == 0:
                        completed = 100 * count / total_samples
                        logger.debug(
                            "[{}] ... {}/{} ({:0.0f}%)".format(
                                self.name, count, total_samples, completed
                            )
                        )

        # build model and run VI
        logger.debug(f"{self.name} building model and running backward induction...")

        N_sa = np.maximum(self.N_sa, 1)
        self.R_hat = self.S_sa / N_sa
        self.P_hat = np.zeros((S, A, S))
        for ss in range(S):
            self.P_hat[:, :, ss] = self.N_sas[:, :, ss] / N_sa

        info = {}
        info["n_samples"] = self.n_samples
        info["total_samples"] = total_samples
        if self.horizon is None:
            assert self.gamma < 1.0, "The discounted setting requires gamma < 1.0"
            self.Q, self.V, n_it = value_iteration(
                self.R_hat, self.P_hat, self.gamma, self.epsilon
            )
            info["n_iterations"] = n_it
            info["precision"] = self.epsilon
        else:
            self.Q, self.V = backward_induction(
                self.R_hat, self.P_hat, self.horizon, self.gamma
            )
            info["n_iterations"] = self.horizon
            info["precision"] = 0.0
        return info

    def policy(self, observation):
        state = observation
        assert self.env.observation_space.contains(state)
        if self.horizon is None:
            return self.Q[state, :].argmax()
        else:
            return self.Q[0, state, :].argmax()



import numpy as np

from rlberry.envs.finite_mdp import FiniteMDP
from rlberry.rendering import RenderInterface2D, Scene, GeometricPrimitive


class Chain(RenderInterface2D, FiniteMDP):
    """
    Simple chain environment.
    Reward 0.05 in initial state, reward 1.0 in final state.

    Parameters
    ----------
    L : int
        length of the chain
    fail_prob : double
        fail probability
    """

    name = "Chain"

    def __init__(self, L=5, fail_prob=0.1):
        assert L >= 2
        self.L = L
        self.fail_prob = fail_prob

        # transition probabilities
        P = np.zeros((L, 2, L))
        for ss in range(L):
            for _ in range(2):
                if ss == 0:
                    P[ss, 0, ss] = 1.0 - fail_prob  # action 0 = don't move
                    P[ss, 1, ss + 1] = 1.0 - fail_prob  # action 1 = right
                    P[ss, 0, ss + 1] = fail_prob
                    P[ss, 1, ss] = fail_prob
                elif ss == L - 1:
                    P[ss, 0, ss - 1] = 1.0 - fail_prob  # action 0 = left
                    P[ss, 1, ss] = 1.0 - fail_prob  # action 1 = don't move
                    P[ss, 0, ss] = fail_prob
                    P[ss, 1, ss - 1] = fail_prob
                else:
                    P[ss, 0, ss - 1] = 1.0 - fail_prob  # action 0 = left
                    P[ss, 1, ss + 1] = 1.0 - fail_prob  # action 1 = right
                    P[ss, 0, ss + 1] = fail_prob
                    P[ss, 1, ss - 1] = fail_prob

                    # mean reward
        S = L
        A = 2
        R = np.zeros((S, A))
        R[L - 1, :] = 1.0
        R[0, :] = 0.05

        # init base classes
        FiniteMDP.__init__(self, R, P, initial_state_distribution=0)
        RenderInterface2D.__init__(self)
        self.reward_range = (0.0, 1.0)

        # rendering info
        self.set_clipping_area((0, L, 0, 1))
        self.set_refresh_interval(100)  # in milliseconds

    def step(self, action):
        assert action in self._actions, "Invalid action!"

        # save state for rendering
        if self.is_render_enabled():
            self.append_state_for_rendering(self.state)

        # take step
        next_state, reward, terminated, truncated, info = self.sample(
            self.state, action
        )

        self.state = next_state
        return next_state, reward, terminated, truncated, info

    #
    # Code for rendering
    #

    def get_background(self):
        """
        Returne a scene (list of shapes) representing the background
        """
        bg = Scene()
        colors = [(0.8, 0.8, 0.8), (0.9, 0.9, 0.9)]
        for ii in range(self.L):
            shape = GeometricPrimitive("QUADS")
            shape.add_vertex((ii, 0))
            shape.add_vertex((ii + 1, 0))
            shape.add_vertex((ii + 1, 1))
            shape.add_vertex((ii, 1))
            shape.set_color(colors[ii % 2])
            bg.add_shape(shape)

        flag = GeometricPrimitive("TRIANGLES")
        flag.set_color((0.0, 0.5, 0.0))
        x = self.L - 0.5
        y = 0.25
        flag.add_vertex((x, y))
        flag.add_vertex((x + 0.25, y + 0.5))
        flag.add_vertex((x - 0.25, y + 0.5))
        bg.add_shape(flag)

        return bg

    def get_scene(self, state):
        """
        Return scene (list of shapes) representing a given state
        """
        scene = Scene()

        agent = GeometricPrimitive("QUADS")
        agent.set_color((0.75, 0.0, 0.5))

        size = 0.25
        x = state + 0.5
        y = 0.5

        agent.add_vertex((x - size / 4.0, y - size))
        agent.add_vertex((x + size / 4.0, y - size))
        agent.add_vertex((x + size / 4.0, y + size))
        agent.add_vertex((x - size / 4.0, y + size))

        agent.add_vertex((x - size, y - size / 4.0))
        agent.add_vertex((x + size, y - size / 4.0))
        agent.add_vertex((x + size, y + size / 4.0))
        agent.add_vertex((x - size, y + size / 4.0))

        scene.add_shape(agent)
        return scene



from rlberry.agents.agent import AgentWithSimplePolicy
from rlberry_scool.agents.dynprog.utils import backward_induction, value_iteration
from rlberry.envs.finite_mdp import FiniteMDP


class ValueIterationAgent(AgentWithSimplePolicy):
    """
    Value iteration for enviroments of type FiniteMDP
    (rlberry.envs.finite.finite_mdp.FiniteMDP)

    Important: the discount gamma is also used if the problem is
    finite horizon, but, in this case, gamma can be set to 1.0.

    Parameters
    -----------
    env : rlberry.envs.finite.finite_mdp.FiniteMDP
        Environment used to fit the agent.
    gamma : double
        Discount factor in [0, 1]
    horizon : int
        Horizon, if the problem is finite-horizon. if None, the discounted
        problem is solved
        default = None
    epsilon : double
        Precision of value iteration, only used in discounted problems
        (when horizon is None).
    **kwargs : Keyword Arguments
         Arguments to be passed to `AgentWithSimplePolicy.__init__(self, env, **kwargs)` (:class:`~rlberry.agents.AgentWithSimplePolicy`).


    """

    name = "ValueIteration"

    def __init__(self, env, gamma=0.95, horizon=None, epsilon=1e-6, **kwargs):
        AgentWithSimplePolicy.__init__(self, env, **kwargs)

        # initialize base class
        assert isinstance(
            self.env, FiniteMDP
        ), "Value iteration requires a FiniteMDP model."
        #

        self.gamma = gamma  # attribute gamma

        self.horizon = horizon
        self.epsilon = epsilon

        # value functions
        self.Q = None
        self.V = None

    def fit(self, budget=None, **kwargs):
        """
        Run value iteration.

        Parameters
        ----------
        budget: None
            Not used. Only defined for compatibility purpose with rlberry.
            Changing `budget` value has no effect.
        **kwargs : Keyword Arguments
            Extra arguments. Not used for this agent.
        """
        del kwargs
        info = {}
        if self.horizon is None:
            assert self.gamma < 1.0, "The discounted setting requires gamma < 1.0"
            self.Q, self.V, n_it = value_iteration(
                self.env.R, self.env.P, self.gamma, self.epsilon
            )
            info["n_iterations"] = n_it
            info["precision"] = self.epsilon
        else:
            self.Q, self.V = backward_induction(
                self.env.R, self.env.P, self.horizon, self.gamma
            )
            info["n_iterations"] = self.horizon
            info["precision"] = 0.0
        return info

    def policy(self, observation):
        state = observation
        if self.horizon is None:
            return self.Q[state, :].argmax()
        else:
            return self.Q[0, state, :].argmax()



import numpy as np
import pytest

import rlberry.seeding as seeding
from rlberry_scool.agents.dynprog import ValueIterationAgent
from rlberry_scool.agents.dynprog.utils import backward_induction
from rlberry_scool.agents.dynprog.utils import backward_induction_in_place
from rlberry_scool.agents.dynprog.utils import backward_induction_sd
from rlberry_scool.agents.dynprog.utils import backward_induction_reward_sd
from rlberry_scool.agents.dynprog.utils import bellman_operator
from rlberry_scool.agents.dynprog.utils import value_iteration
from rlberry.envs.finite_mdp import FiniteMDP

_rng = seeding.Seeder(123).rng


def get_random_mdp(S, A):
    R = _rng.uniform(0.0, 1.0, (S, A))
    P = _rng.uniform(0.0, 1.0, (S, A, S))
    for ss in range(S):
        for aa in range(A):
            P[ss, aa, :] /= P[ss, aa, :].sum()
    return R, P


@pytest.mark.parametrize(
    "gamma, S, A",
    [
        (0.001, 2, 1),
        (0.25, 2, 1),
        (0.5, 2, 1),
        (0.75, 2, 1),
        (0.999, 2, 1),
        (0.001, 4, 2),
        (0.25, 4, 2),
        (0.5, 4, 2),
        (0.75, 4, 2),
        (0.999, 4, 2),
        (0.001, 20, 4),
        (0.25, 20, 4),
        (0.5, 20, 4),
        (0.75, 20, 4),
        (0.999, 20, 4),
    ],
)
def test_bellman_operator_monotonicity_and_contraction(gamma, S, A):
    rng = seeding.Seeder(123).rng
    vmax = 1.0 / (1.0 - gamma)
    for _ in range(10):
        # generate random MDP
        R, P = get_random_mdp(S, A)

        # generate random Q functions
        Q0 = rng.uniform(-vmax, vmax, (S, A))
        Q1 = rng.uniform(-vmax, vmax, (S, A))
        # apply Bellman operator
        TQ0 = bellman_operator(Q0, R, P, gamma)
        TQ1 = bellman_operator(Q1, R, P, gamma)

        # test contraction
        norm_tq = np.abs(TQ1 - TQ0).max()
        norm_q = np.abs(Q1 - Q0).max()
        assert norm_tq <= gamma * norm_q

        # test monotonicity
        Q2 = rng.uniform(-vmax / 2, vmax / 2, (S, A))
        Q3 = Q2 + rng.uniform(0.0, vmax / 2, (S, A))
        TQ2 = bellman_operator(Q2, R, P, gamma)
        TQ3 = bellman_operator(Q3, R, P, gamma)
        assert np.greater(TQ2, TQ3).sum() == 0


@pytest.mark.parametrize(
    "gamma, S, A",
    [(0.01, 10, 4), (0.25, 10, 4), (0.5, 10, 4), (0.75, 10, 4), (0.99, 10, 4)],
)
def test_value_iteration(gamma, S, A):
    for epsilon in np.logspace(-1, -6, num=5):
        for sim in range(5):
            # generate random MDP
            R, P = get_random_mdp(S, A)

            # run value iteration
            Q, V, n_it = value_iteration(R, P, gamma, epsilon)
            # check precision
            TQ = bellman_operator(Q, R, P, gamma)
            assert np.abs(TQ - Q).max() <= epsilon


@pytest.mark.parametrize("horizon, S, A", [(10, 5, 4), (20, 10, 4)])
def test_backward_induction(horizon, S, A):
    for sim in range(5):
        # generate random MDP
        R, P = get_random_mdp(S, A)

        # run backward induction
        Q, V = backward_induction(R, P, horizon)

        assert Q.max() <= horizon
        assert V.max() <= horizon

        # run backward with clipping V to 1.0
        Q, V = backward_induction(R, P, horizon, vmax=1.0)
        assert V.max() <= 1.0

        # run bacward induction in place
        Q2 = np.zeros((horizon, S, A))
        V2 = np.zeros((horizon, S))
        backward_induction_in_place(Q2, V2, R, P, horizon, vmax=1.0)
        assert np.array_equal(Q, Q2)
        assert np.array_equal(V, V2)


@pytest.mark.parametrize("horizon, S, A", [(10, 5, 4), (20, 10, 4)])
def test_backward_induction_sd(horizon, S, A):
    """
    Test stage-dependent MDPs
    """
    for sim in range(5):
        # generate random MDP
        Rstat, Pstat = get_random_mdp(S, A)
        R = np.zeros((horizon, S, A))
        P = np.zeros((horizon, S, A, S))
        for ii in range(horizon):
            R[ii, :, :] = Rstat
            P[ii, :, :, :] = Pstat

        # run backward induction in stationary MDP
        Qstat, Vstat = backward_induction(Rstat, Pstat, horizon)

        # run backward induction in stage-dependent MDP
        Q = np.zeros((horizon, S, A))
        V = np.zeros((horizon, S))
        backward_induction_sd(Q, V, R, P)

        # run backward induction with stage-dependent rewards
        Q2 = np.zeros((horizon, S, A))
        V2 = np.zeros((horizon, S))
        backward_induction_reward_sd(Q2, V2, R, Pstat)

        assert np.array_equal(Q, Qstat)
        assert np.array_equal(V, Vstat)
        assert np.array_equal(Q2, Qstat)
        assert np.array_equal(V2, Vstat)


@pytest.mark.parametrize("horizon, gamma, S, A", [(None, 0.5, 10, 4), (10, 1.0, 10, 4)])
def test_value_iteration_agent(horizon, gamma, S, A):
    for sim in range(5):
        # generate random MDP
        R, P = get_random_mdp(S, A)
        # create env and agent
        env = FiniteMDP(R, P)
        agent = ValueIterationAgent(env, gamma=gamma, horizon=horizon)
        # run
        agent.fit()



from rlberry_scool.envs.finite import Chain, GridWorld
from rlberry_scool.agents.dynprog import ValueIterationAgent
import tempfile
import os


def test_chain_rendering():
    env = Chain(10, 0.3)
    env.enable_rendering()
    for tt in range(5):
        env.step(env.action_space.sample())
    with tempfile.TemporaryDirectory() as tmpdirname:
        saving_path = tmpdirname + "/test_gif.gif"
        env.save_gif(saving_path)
        assert os.path.isfile(saving_path)
        try:
            os.remove(saving_path)
        except Exception:
            pass


def test_gridworld_rendering():
    env = GridWorld(7, 10, walls=((2, 2), (3, 3)))

    agent = ValueIterationAgent(env, gamma=0.95)
    info = agent.fit()
    print(info)

    env.enable_rendering()
    observation, info = env.reset()
    for tt in range(50):
        action = agent.policy(observation)
        observation, reward, terminated, truncated, info = env.step(action)
        done = terminated or truncated
        if done:
            # Warning: this will never happen in the present case because there is no terminal state.
            # See the doc of GridWorld for more informations on the default parameters of GridWorld.
            break

    with tempfile.TemporaryDirectory() as tmpdirname:
        saving_path = tmpdirname + "/test_gif.gif"
        env.save_gif(saving_path)
        assert os.path.isfile(saving_path)
        try:
            os.remove(saving_path)
        except Exception:
            pass



import numpy as np
from rlberry.utils.jit_setup import numba_jit


@numba_jit
def backward_induction(R, P, horizon, gamma=1.0, vmax=np.inf):
    """Backward induction to compute Q and V functions in the finite horizon
    setting.

    Parameters
    ----------
    R : numpy.ndarray
        array of shape (S, A) contaning the rewards, where S is the number
        of states and A is the number of actions
    P : numpy.ndarray
        array of shape (S, A, S) such that P[s,a,ns] is the probability of
        arriving at ns by taking action a in state s.
    horizon : int
        problem horizon
    gamma : double, default: 1.0
        discount factor
    vmax : double, default: np.inf
        maximum possible value in V

    Returns
    --------
    tuple (Q, V) containing the Q and V functions, of shapes (horizon, S, A)
    and (horizon, S), respectively.
    """
    S, A = R.shape
    V = np.zeros((horizon, S))
    Q = np.zeros((horizon, S, A))
    for hh in range(horizon - 1, -1, -1):
        for ss in range(S):
            max_q = -np.inf
            for aa in range(A):
                q_aa = R[ss, aa]
                if hh < horizon - 1:
                    # not using .dot instead of loop to avoid scipy dependency
                    # (numba seems to require scipy for linear
                    # algebra operations in numpy)
                    for ns in range(S):
                        q_aa += gamma * P[ss, aa, ns] * V[hh + 1, ns]
                if q_aa > max_q:
                    max_q = q_aa
                Q[hh, ss, aa] = q_aa
            V[hh, ss] = max_q
            if V[hh, ss] > vmax:
                V[hh, ss] = vmax
    return Q, V


@numba_jit
def backward_induction_reward_sd(Q, V, R, P, gamma=1.0, vmax=np.inf):
    """
    Backward induction to compute Q and V functions in
    the finite horizon setting.

    Assumes R is stage-dependent, but P is stage-independent.

    Takes as input the arrays where to store Q and V.

    Parameters
    ----------
    Q:  numpy.ndarray
        array of shape (horizon, S, A) where to store the Q function
    V:  numpy.ndarray
        array of shape (horizon, S) where to store the V function
    R : numpy.ndarray
        array of shape (horizon, S, A) contaning the rewards, where S is the number
        of states and A is the number of actions
    P : numpy.ndarray
        array of shape (S, A, S) such that P[s,a,ns] is the probability of
        arriving at ns by taking action a in state s.
    horizon : int
        problem horizon
    gamma : double
        discount factor, default = 1.0
    vmax : double
        maximum possible value in V
        default = np.inf
    """
    H, S, A = R.shape
    horizon = H
    for hh in range(horizon - 1, -1, -1):
        for ss in range(S):
            max_q = -np.inf
            for aa in range(A):
                q_aa = R[hh, ss, aa]
                if hh < horizon - 1:
                    # not using .dot instead of loop to avoid scipy dependency
                    # (numba seems to require scipy for linear algebra
                    #  operations in numpy)
                    for ns in range(S):
                        q_aa += gamma * P[ss, aa, ns] * V[hh + 1, ns]
                if q_aa > max_q:
                    max_q = q_aa
                Q[hh, ss, aa] = q_aa
            V[hh, ss] = max_q
            if V[hh, ss] > vmax:
                V[hh, ss] = vmax


@numba_jit
def backward_induction_in_place(Q, V, R, P, horizon, gamma=1.0, vmax=np.inf):
    """
    Backward induction to compute Q and V functions in
    the finite horizon setting.
    Takes as input the arrays where to store Q and V.

    Parameters
    ----------
    Q:  numpy.ndarray
        array of shape (horizon, S, A) where to store the Q function
    V:  numpy.ndarray
        array of shape (horizon, S) where to store the V function
    R : numpy.ndarray
        array of shape (S, A) contaning the rewards, where S is the number
        of states and A is the number of actions
    P : numpy.ndarray
        array of shape (S, A, S) such that P[s,a,ns] is the probability of
        arriving at ns by taking action a in state s.
    horizon : int
        problem horizon
    gamma : double
        discount factor, default = 1.0
    vmax : double
        maximum possible value in V
        default = np.inf
    """
    S, A = R.shape
    for hh in range(horizon - 1, -1, -1):
        for ss in range(S):
            max_q = -np.inf
            for aa in range(A):
                q_aa = R[ss, aa]
                if hh < horizon - 1:
                    # not using .dot instead of loop to avoid scipy dependency
                    # (numba seems to require scipy for linear algebra
                    #  operations in numpy)
                    for ns in range(S):
                        q_aa += gamma * P[ss, aa, ns] * V[hh + 1, ns]
                if q_aa > max_q:
                    max_q = q_aa
                Q[hh, ss, aa] = q_aa
            V[hh, ss] = max_q
            if V[hh, ss] > vmax:
                V[hh, ss] = vmax


@numba_jit
def backward_induction_sd(Q, V, R, P, gamma=1.0, vmax=np.inf):
    """
    In-place implementation of backward induction to compute Q and V functions
    in the finite horizon setting.

    Assumes R and P are stage-dependent.

    Parameters
    ----------
    Q:  numpy.ndarray
        array of shape (H, S, A) where to store the Q function
    V:  numpy.ndarray
        array of shape (H, S) where to store the V function
    R : numpy.ndarray
        array of shape (H, S, A) contaning the rewards, where S is the number
        of states and A is the number of actions
    P : numpy.ndarray
        array of shape (H, S, A, S) such that P[h, s, a, ns] is the probability of
        arriving at ns by taking action a in state s at stage h.
    gamma : double, default: 1.0
        discount factor
    vmax : double, default: np.inf
        maximum possible value in V

    """
    H, S, A = R.shape
    for hh in range(H - 1, -1, -1):
        for ss in range(S):
            max_q = -np.inf
            for aa in range(A):
                q_aa = R[hh, ss, aa]
                if hh < H - 1:
                    # not using .dot instead of loop to avoid scipy dependency
                    # (numba seems to require scipy for linear
                    # algebra operations in numpy)
                    for ns in range(S):
                        q_aa += gamma * P[hh, ss, aa, ns] * V[hh + 1, ns]
                if q_aa > max_q:
                    max_q = q_aa
                Q[hh, ss, aa] = q_aa
            V[hh, ss] = max_q
            # clip V
            if V[hh, ss] > vmax:
                V[hh, ss] = vmax


@numba_jit
def value_iteration(R, P, gamma, epsilon=1e-6):
    """
    Value iteration for discounted problems.

    Parameters
    ----------
    R : numpy.ndarray
        array of shape (S, A) contaning the rewards, where S is the number
        of states and A is the number of actions
    P : numpy.ndarray
        array of shape (S, A, S) such that P[s,a,ns] is the probability of
        arriving at ns by taking action a in state s.
    gamma : double
        discount factor
    epsilon : double
        precision

    Returns
    --------
    tuple (Q, V, n_it) containing the epsilon-optimal Q and V functions,
    of shapes (S, A) and (S,), respectively, and n_it, the number of iterations
    """
    S, A = R.shape
    Q = np.zeros((S, A))
    Q_aux = np.full((S, A), np.inf)
    n_it = 0
    while np.abs(Q - Q_aux).max() > epsilon:
        Q_aux = Q
        Q = bellman_operator(Q, R, P, gamma)
        n_it += 1
    V = np.zeros(S)
    # numba does not support np.max(Q, axis=1)
    for ss in range(S):
        V[ss] = Q[ss, :].max()
    return Q, V, n_it


@numba_jit
def bellman_operator(Q, R, P, gamma):
    """
    Bellman optimality operator for Q functions

    Parameters
    ----------
    Q : numpy.ndarray
        array of shape (S, A) containing the Q function to which apply
        the operator
    R : numpy.ndarray
        array of shape (S, A) contaning the rewards, where S is the number
        of states and A is the number of actions
    P : numpy.ndarray
        array of shape (S, A, S) such that P[s,a,ns] is the probability of
        arriving at ns by taking action a in state s.
    gamma : double
        discount factor

    Returns
    --------
    TQ, array of shape (S, A) containing the result of the Bellman operator
    applied to the input Q
    """
    S, A = Q.shape
    TQ = np.zeros((S, A))
    V = np.zeros(S)
    # numba does not support np.max(Q, axis=1)
    for ss in range(S):
        V[ss] = Q[ss, :].max()
    #
    for ss in range(S):
        for aa in range(A):
            TQ[ss, aa] = R[ss, aa]
            for ns in range(S):
                TQ[ss, aa] += gamma * P[ss, aa, ns] * V[ns]
    return TQ



from .gridworld import GridWorld
from .chain import Chain
from .discrete_mountain_car import get_discrete_mountain_car_env

from gymnasium.envs.registration import register

register(
    id="rlberry_scool/chain",
    entry_point="rlberry_scool.envs.finite:Chain",
    max_episode_steps=1000,
)



import numpy as np
import pytest
import rlberry.seeding as seeding

from copy import deepcopy
from rlberry_scool.envs.finite import Chain
from rlberry_scool.envs.finite import GridWorld

classes = [
    GridWorld,
    Chain,
]


def get_env_trajectory(env, horizon):
    states = []
    ss, info = env.reset()
    for ii in range(horizon):
        states.append(ss)
        ss, _, _, _, _ = env.step(env.action_space.sample())
    return states


def compare_trajectories(traj1, traj2):
    """
    returns true if trajectories are equal
    """
    for ss1, ss2 in zip(traj1, traj2):
        if not np.array_equal(ss1, ss2):
            return False
    return True


@pytest.mark.parametrize("ModelClass", classes)
def test_env_seeding(ModelClass):
    env1 = ModelClass()
    seeder1 = seeding.Seeder(123)
    env1.reseed(seeder1)

    env2 = ModelClass()
    seeder2 = seeder1.spawn()
    env2.reseed(seeder2)

    env3 = ModelClass()
    seeder3 = seeding.Seeder(123)
    env3.reseed(seeder3)

    env4 = ModelClass()
    seeder4 = seeding.Seeder(123)
    env4.reseed(seeder4)

    env5 = ModelClass()
    env5.reseed(
        seeder1
    )  # same seeder as env1, but different trajectories. This is expected.

    seeding.safe_reseed(env4, seeder4)

    if deepcopy(env1).is_online():
        traj1 = get_env_trajectory(env1, 500)
        traj2 = get_env_trajectory(env2, 500)
        traj3 = get_env_trajectory(env3, 500)
        traj4 = get_env_trajectory(env4, 500)
        traj5 = get_env_trajectory(env5, 500)

        assert not compare_trajectories(traj1, traj2)
        assert compare_trajectories(traj1, traj3)
        assert not compare_trajectories(traj3, traj4)
        assert not compare_trajectories(traj1, traj5)


@pytest.mark.parametrize("ModelClass", classes)
def test_copy_reseeding(ModelClass):
    seeder = seeding.Seeder(123)
    env = ModelClass()
    env.reseed(seeder)

    c_env = deepcopy(env)
    c_env.reseed()

    if deepcopy(env).is_online():
        traj1 = get_env_trajectory(env, 500)
        traj2 = get_env_trajectory(c_env, 500)
        assert not compare_trajectories(traj1, traj2)



from typing import Optional, Literal
import numpy as np
from gymnasium import spaces
from scipy.special import softmax

from rlberry import types
from rlberry.agents import AgentWithSimplePolicy


class SARSAAgent(AgentWithSimplePolicy):
    """SARSA Agent.

    Parameters
    ----------
    env: :class:`~rlberry.types.Env`
        Environment with discrete states and actions.
    gamma: float, default = 0.99
        Discount factor.
    alpha: float, default = 0.1
        Learning rate.
    exploration_type: {"epsilon", "boltzmann"}, default: None
        If "epsilon": Epsilon-Greedy exploration.
        If "boltzmann": Boltzmann exploration.
        If None: No exploration.
    exploration_rate: float, default: None
        epsilon parameter for Epsilon-Greedy exploration or tau parameter for Boltzmann exploration.
    **kwargs : Keyword Arguments
         Arguments to be passed to `AgentWithSimplePolicy.__init__(self, env, **kwargs)` (:class:`~rlberry.agents.AgentWithSimplePolicy`).


    Attributes
    ----------
    Q : ndarray
        2D array that stores the estimation ofexpected rewards for state-action pairs.
    Examples
    --------
    >>> from rlberry.envs import GridWorld
    >>>
    >>> env = GridWorld(walls=(), nrows=5, ncols=5)
    >>> agent = SARSAAgent()
    >>> agent.fit(budget=1000)
    >>> agent.policy(env.observation_space.sample())
    >>> agent.reset()
    """

    def __init__(
        self,
        env: types.Env,
        gamma: float = 0.99,
        alpha: float = 0.1,
        exploration_type: Optional[Literal["epsilon", "boltzmann"]] = None,
        exploration_rate: Optional[float] = None,
        **kwargs
    ):
        # init base class
        AgentWithSimplePolicy.__init__(self, env, **kwargs)

        self.gamma = gamma
        self.alpha = alpha
        self.exploration_type = exploration_type
        self.exploration_rate = exploration_rate
        # check environment
        assert isinstance(self.env.observation_space, spaces.Discrete)
        assert isinstance(self.env.action_space, spaces.Discrete)

        # check exploration type
        if self.exploration_type is not None:
            assert (
                exploration_type == "epsilon" or "boltzmann"
            ) and exploration_rate is not None

        self.Q = np.zeros((self.env.observation_space.n, self.env.action_space.n))

    def reset(self, **kwargs):
        self.Q.fill(0)

    def policy(self, observation):
        return self.Q[observation].argmax()

    def get_action(self, observation):
        if (
            self.exploration_type == "epsilon"
            and np.random.random() <= self.exploration_rate
        ):
            return np.random.choice(self.env.action_space.n)
        elif self.exploration_type == "boltzmann":
            return np.random.choice(
                self.env.action_space.n,
                p=softmax(self.exploration_rate * self.Q[observation]),
            )
        else:
            return self.Q[observation].argmax()

    def fit(self, budget: int, **kwargs):
        """
        Train the agent using the provided environment.
        Parameters
        ----------
        budget: int
            number of Q updates.
        **kwargs : Keyword Arguments
            Extra arguments. Not used for this agent.
        """
        del kwargs
        observation, info = self.env.reset()
        episode_rewards = 0
        for i in range(budget):
            action = self.get_action(observation)
            next_observation, reward, terminated, truncated, info = self.env.step(
                action
            )
            done = terminated or truncated
            episode_rewards += reward
            if self.writer is not None:
                self.writer.add_scalar("episode_rewards", episode_rewards, i)
            if done:
                self.Q[observation, action] = reward
            else:
                next_action = self.get_action(next_observation)
                self.Q[observation, action] = self.Q[
                    observation, action
                ] + self.alpha * (
                    reward
                    + self.gamma * self.Q[next_observation, next_action]
                    - self.Q[observation, action]
                )
            observation = next_observation
            if done:
                observation, info = self.env.reset()
                episode_rewards = 0



import numpy as np
from rlberry.agents import AgentWithSimplePolicy
from gymnasium.spaces import Discrete
from rlberry.utils.jit_setup import numba_jit

import rlberry

logger = rlberry.logger


@numba_jit
def run_lsvi_jit(
    dim,
    horizon,
    bonus_factor,
    lambda_mat_inv,
    reward_hist,
    gamma,
    feat_hist,
    n_actions,
    feat_ns_all_actions,
    v_max,
    total_time_steps,
):
    """
    Jit version of Least-Squares Value Iteration.

    Parameters
    ----------
    dim : int
        Dimension of the features
    horiton : int

    bonus_factor : int

    lambda_mat_inv : numpy array (dim, dim)
        Inverse of the design matrix

    reward_hist : numpy array (time,)

    gamma : double

    feat_hist : numpy array (time, dim)

    n_actions : int

    feat_ns_all_actions : numpy array (time, n_actions, dim)
        History of next state features for all actions

    vmax : double
        Maximum value of the value function

    total_time_steps : int
        Current step count
    """
    # run value iteration
    q_w = np.zeros((horizon + 1, dim))
    for hh in range(horizon - 1, -1, -1):
        T = total_time_steps
        b = np.zeros(dim)
        for tt in range(T):
            # compute q function at next state, q_ns
            q_ns = np.zeros(n_actions)
            for aa in range(n_actions):
                #
                feat_ns_aa = feat_ns_all_actions[tt, aa, :]
                inverse_counts = feat_ns_aa.dot(lambda_mat_inv.T.dot(feat_ns_aa))
                bonus = bonus_factor * np.sqrt(
                    inverse_counts
                ) + v_max * inverse_counts * (bonus_factor > 0.0)
                #
                q_ns[aa] = feat_ns_aa.dot(q_w[hh + 1, :]) + bonus
                q_ns[aa] = min(q_ns[aa], v_max)

            # compute regretion targets
            target = reward_hist[tt] + gamma * q_ns.max()
            feat = feat_hist[tt, :]
            b = b + target * feat

        # solve M x = b, where x = q_w, and M = self.lambda_mat
        q_w[hh, :] = lambda_mat_inv.T @ b
    return q_w


class LSVIUCBAgent(AgentWithSimplePolicy):
    """
    A version of Least-Squares Value Iteration with UCB (LSVI-UCB),
    proposed by Jin et al. (2020).

    If bonus_scale_factor is 0.0, performs random exploration.

    Notes
    -----
    The computation of exploration bonuses was adapted to match the "simplified Bernstein"
    bonuses that works well empirically for UCBVI in the tabular case.

    The transition probabilities are assumed to be *independent* of the timestep h.

    Parameters
    ----------
    env : Model
        Online model of an environment.
    horizon : int
        Maximum length of each episode.
    feature_map_fn : function(env, kwargs)
        Function that returns a feature map instance
        (rlberry.agents.features.FeatureMap class).
    feature_map_kwargs:
        kwargs for feature_map_fn
    gamma : double
        Discount factor.
    bonus_scale_factor : double
        Constant by which to multiply the exploration bonus.
    reg_factor : double
        Linear regression regularization factor.
    **kwargs : Keyword Arguments
         Arguments to be passed to `AgentWithSimplePolicy.__init__(self, env, **kwargs)` (:class:`~rlberry.agents.AgentWithSimplePolicy`).


    References
    ----------
    Jin, C., Yang, Z., Wang, Z., & Jordan, M. I. (2020, July).
    Provably efficient reinforcement learning with linear
    function approximation. In Conference on Learning Theory (pp. 2137-2143).
    """

    name = "LSVI-UCB"

    def __init__(
        self,
        env,
        horizon,
        feature_map_fn,
        feature_map_kwargs=None,
        gamma=0.99,
        bonus_scale_factor=1.0,
        reg_factor=0.1,
        **kwargs
    ):
        AgentWithSimplePolicy.__init__(self, env, **kwargs)

        self.n_episodes = None
        self.horizon = horizon
        self.gamma = gamma
        self.bonus_scale_factor = bonus_scale_factor
        self.reg_factor = reg_factor
        feature_map_kwargs = feature_map_kwargs or {}
        self.feature_map = feature_map_fn(self.env, **feature_map_kwargs)

        #
        if self.bonus_scale_factor == 0.0:
            self.name = "LSVI-Random-Expl"

        # maximum value
        r_range = self.env.reward_range[1] - self.env.reward_range[0]
        if r_range == np.inf:
            logger.warning(
                "{}: Reward range is infinity. ".format(self.name) + "Clipping it to 1."
            )
            r_range = 1.0

        if self.gamma == 1.0:
            self.v_max = r_range * horizon
        else:
            self.v_max = (
                r_range
                * (1.0 - np.power(self.gamma, self.horizon))
                / (1.0 - self.gamma)
            )

        #
        assert isinstance(
            self.env.action_space, Discrete
        ), "LSVI-UCB requires discrete actions."

        #
        assert len(self.feature_map.shape) == 1
        self.dim = self.feature_map.shape[0]

        # attributes initialized in reset()
        self.episode = None
        self.lambda_mat = None  # lambda matrix
        self.lambda_mat_inv = None  # inverse of lambda matrix
        self.w_vec = None  # vector representation of Q
        self.w_policy = None  # representation of Q for final policy
        self.reward_hist = None  # reward history
        self.state_hist = None  # state history
        self.action_hist = None  # action history
        self.nstate_hist = None  # next state history

        self.feat_hist = None  # feature history
        self.feat_ns_all_actions = None  # next state features for all actions
        #

        # aux variables (init in reset() too)
        self._rewards = None

    def reset(self):
        self.episode = 0
        self.total_time_steps = 0
        self.lambda_mat = self.reg_factor * np.eye(self.dim)
        self.lambda_mat_inv = (1.0 / self.reg_factor) * np.eye(self.dim)
        self.w_vec = np.zeros((self.horizon + 1, self.dim))
        self.reward_hist = np.zeros(self.n_episodes * self.horizon)
        self.state_hist = []
        self.action_hist = []
        self.nstate_hist = []
        # episode rewards
        self._rewards = np.zeros(self.n_episodes)
        #
        self.feat_hist = np.zeros((self.n_episodes * self.horizon, self.dim))
        self.feat_ns_all_actions = np.zeros(
            (self.n_episodes * self.horizon, self.env.action_space.n, self.dim)
        )
        #
        self.w_policy = None

    def fit(self, budget, **kwargs):
        """
        Train the agent using the provided environment.

        Parameters
        ----------
        budget: int
            number of episodes. Each episode runs for self.horizon unless it
            enconters a terminal state in which case it stops early.
            Warning: Calling fit() more than once will reset the algorithm
            (to realocate memory according to the number of episodes)
        **kwargs : Keyword Arguments
            Extra arguments. Not used for this agent.
        """
        del kwargs

        # Allocate memory according to budget.
        # TODO: avoid the need to reset() the algorithm if fit() is called again.
        if self.n_episodes is not None:
            logger.warning(
                "[LSVI-UCB]: Calling fit() more than once will reset the algorithm"
                + " (to realocate memory according to the number of episodes)."
            )
        self.n_episodes = budget
        self.reset()

        for ep in range(self.n_episodes):
            self.run_episode()
            if self.bonus_scale_factor > 0.0 or ep == self.n_episodes - 1:
                # update Q function representation
                self.w_vec = self._run_lsvi(self.bonus_scale_factor)

        self.w_policy = self._run_lsvi(bonus_factor=0.0)[0, :]

    def policy(self, observation):
        q_w = self.w_policy
        assert q_w is not None
        #
        q_vec = self._compute_q_vec(q_w, observation, 0.0)
        return q_vec.argmax()

    def _optimistic_policy(self, observation, hh):
        q_w = self.w_vec[hh, :]
        q_vec = self._compute_q_vec(q_w, observation, self.bonus_scale_factor)
        return q_vec.argmax()

    def run_episode(self):
        observation, info = self.env.reset()
        episode_rewards = 0
        for hh in range(self.horizon):
            if self.bonus_scale_factor == 0.0:
                action = self.env.action_space.sample()
            else:
                action = self._optimistic_policy(observation, hh)

            next_observation, reward, terminated, truncated, info = self.env.step(
                action
            )
            done = terminated or truncated

            feat = self.feature_map.map(observation, action)
            outer_prod = np.outer(feat, feat)
            inv = self.lambda_mat_inv

            #
            self.lambda_mat += np.outer(feat, feat)
            # update inverse
            self.lambda_mat_inv -= (inv @ outer_prod @ inv) / (1 + feat @ inv.T @ feat)

            # update history
            self.reward_hist[self.total_time_steps] = reward
            self.state_hist.append(observation)
            self.action_hist.append(action)
            self.nstate_hist.append(next_observation)

            #
            tt = self.total_time_steps
            self.feat_hist[tt, :] = self.feature_map.map(observation, action)
            for aa in range(self.env.action_space.n):
                self.feat_ns_all_actions[tt, aa, :] = self.feature_map.map(
                    next_observation, aa
                )

            # increments
            self.total_time_steps += 1
            episode_rewards += reward

            #
            observation = next_observation
            if done:
                break

        # store data
        self._rewards[self.episode] = episode_rewards

        # update ep
        self.episode += 1

        #
        if self.writer is not None:
            self.writer.add_scalar("episode_rewards", episode_rewards, self.episode)

        return episode_rewards

    def _compute_q(self, q_w, state, action, bonus_factor):
        """q_w is the vector representation of the Q function."""
        feat = self.feature_map.map(state, action)
        inverse_counts = feat @ (self.lambda_mat_inv.T @ feat)
        bonus = bonus_factor * np.sqrt(inverse_counts) + self.v_max * inverse_counts * (
            bonus_factor > 0.0
        )
        q = feat.dot(q_w) + bonus
        return q

    def _compute_q_vec(self, q_w, state, bonus_factor):
        A = self.env.action_space.n
        q_vec = np.zeros(A)
        for aa in range(A):
            # q_vec[aa] = self._compute_q(q_w, state, aa, bonus_factor)
            feat = self.feature_map.map(state, aa)
            inverse_counts = feat @ (self.lambda_mat_inv.T @ feat)
            bonus = bonus_factor * np.sqrt(
                inverse_counts
            ) + self.v_max * inverse_counts * (bonus_factor > 0.0)
            q_vec[aa] = feat.dot(q_w) + bonus
            # q_vec[aa] = min(q_vec[aa], self.v_max)   # !!!!!!!!!
        return q_vec

    def _run_lsvi(self, bonus_factor):
        # run value iteration
        q_w = run_lsvi_jit(
            self.dim,
            self.horizon,
            bonus_factor,
            self.lambda_mat_inv,
            self.reward_hist,
            self.gamma,
            self.feat_hist,
            self.env.action_space.n,
            self.feat_ns_all_actions,
            self.v_max,
            self.total_time_steps,
        )
        return q_w



#
#  Code to visualize the environments.
#

import base64
from pyvirtualdisplay import Display
from IPython import display as ipythondisplay

# from IPython.display import clear_output
from pathlib import Path


def show_video(filename=None, directory="./videos"):
    """
    Either show all videos in a directory (if filename is None) or
    show video corresponding to filename.
    """
    html = []
    if filename is not None:
        files = Path("./").glob(filename)
    else:
        files = Path(directory).glob("*.mp4")
    for mp4 in files:
        video_b64 = base64.b64encode(mp4.read_bytes())
        html.append(
            """<video alt="{}" autoplay
                      loop controls style="height: 400px;">
                    <source src="data:video/mp4;base64,{}" type="video/mp4" />
                 </video>""".format(
                mp4, video_b64.decode("ascii")
            )
        )
    ipythondisplay.display(ipythondisplay.HTML(data="<br>".join(html)))


display = Display(visible=0, size=(1400, 900))
display.start()



import numpy as np
import pytest

from rlberry.seeding import Seeder
from rlberry_scool.agents.mbqvi import MBQVIAgent
from rlberry.envs.finite_mdp import FiniteMDP


@pytest.mark.parametrize("S, A", [(5, 2), (10, 4)])
def test_mbqvi(S, A):
    rng = Seeder(123).rng

    for sim in range(5):
        # generate random MDP with deterministic transitions
        R = rng.uniform(0.0, 1.0, (S, A))
        P = np.zeros((S, A, S))
        for ss in range(S):
            for aa in range(A):
                ns = rng.integers(0, S)
                P[ss, aa, ns] = 1

        # run MBQVI and check exactness of estimators
        env = FiniteMDP(R, P)
        agent = MBQVIAgent(env, n_samples=1)
        agent.fit()
        assert np.abs(R - agent.R_hat).max() < 1e-16
        assert np.abs(P - agent.P_hat).max() < 1e-16



import numpy as np
import pytest
from rlberry_scool.agents.features import FeatureMap
from rlberry_scool.agents.linear.lsvi_ucb import LSVIUCBAgent
from rlberry_scool.agents.dynprog import ValueIterationAgent
from rlberry_scool.envs.finite import GridWorld


class OneHotFeatureMap(FeatureMap):
    def __init__(self, S, A):
        self.S = S
        self.A = A
        self.shape = (S * A,)

    def map(self, observation, action):
        feat = np.zeros((self.S, self.A))
        feat[observation, action] = 1.0
        return feat.flatten()


class RandomFeatMap(FeatureMap):
    def __init__(self, S, A):
        self.feat_mat = np.random.randn(S, A, 10)
        self.shape = (10,)

    def map(self, observation, action):
        feat = self.feat_mat[observation, action, :]
        return feat.copy()


@pytest.mark.parametrize("FeatMapClass", [OneHotFeatureMap, RandomFeatMap])
def test_lsvi_ucb_matrix_inversion(FeatMapClass):
    env = GridWorld(nrows=3, ncols=3, walls=())
    env.reseed(123)

    def feature_map_fn(_env):
        return FeatMapClass(_env.observation_space.n, _env.action_space.n)

    reg_factor = 0.1
    agent = LSVIUCBAgent(
        env, feature_map_fn=feature_map_fn, horizon=10, reg_factor=reg_factor
    )
    agent.reseed(123)
    agent.fit(budget=50)
    assert np.allclose(np.linalg.inv(agent.lambda_mat), agent.lambda_mat_inv)
    assert agent.episode == 50
    agent.policy(env.observation_space.sample())

    # Check counts
    if FeatMapClass != OneHotFeatureMap:
        return

    S = env.observation_space.n
    A = env.action_space.n
    N_sa = np.zeros((S, A))
    for state, action in zip(agent.state_hist, agent.action_hist):
        N_sa[state, action] += 1.0

    assert np.allclose(
        agent.lambda_mat_inv.diagonal(), 1.0 / (N_sa.flatten() + reg_factor)
    )

    for ss in range(S):
        for aa in range(A):
            feat = agent.feature_map.map(ss, aa)
            assert np.allclose(
                feat @ (agent.lambda_mat_inv.T @ feat),
                1.0 / (N_sa[ss, aa] + reg_factor),
            )


def test_lsvi_without_bonus():
    def lsvi_debug_gather_data(agent):
        """
        Function to gather data sampling uniformly
        states and actions
        """
        N = agent.n_episodes * agent.horizon
        count = 0
        while count < N:
            state = agent.env.observation_space.sample()
            action = agent.env.action_space.sample()
            next_state, reward, terminated, truncated, info = agent.env.sample(
                state, action
            )
            done = terminated or truncated
            #
            #
            feat = agent.feature_map.map(state, action)
            outer_prod = np.outer(feat, feat)
            inv = agent.lambda_mat_inv

            #
            agent.lambda_mat += np.outer(feat, feat)
            # update inverse
            agent.lambda_mat_inv -= (inv @ outer_prod @ inv) / (1 + feat @ inv.T @ feat)

            # update history
            agent.reward_hist[count] = reward
            agent.state_hist.append(state)
            agent.action_hist.append(action)
            agent.nstate_hist.append(next_state)

            #
            tt = agent.total_time_steps
            agent.feat_hist[tt, :] = agent.feature_map.map(state, action)
            for aa in range(agent.env.action_space.n):
                agent.feat_ns_all_actions[tt, aa, :] = agent.feature_map.map(
                    next_state, aa
                )

            # increments
            agent.total_time_steps += 1
            count += 1

    env = GridWorld(nrows=2, ncols=2, walls=(), success_probability=0.95)
    env.reseed(123)

    def feature_map_fn(_env):
        return OneHotFeatureMap(_env.observation_space.n, _env.action_space.n)

    agent = LSVIUCBAgent(
        env, feature_map_fn=feature_map_fn, horizon=20, gamma=0.99, reg_factor=1e-5
    )
    agent.reseed(123)
    agent.n_episodes = 100
    agent.reset()

    lsvi_debug_gather_data(agent)
    # estimated Q
    S = env.observation_space.n
    Q_est = agent._run_lsvi(bonus_factor=0.0)[0, :].reshape((S, -1))

    # near optimal Q
    agent_opt = ValueIterationAgent(env, gamma=0.99, horizon=20)
    agent_opt.fit()
    Q = agent_opt.Q[0, :, :]

    print(Q)
    print("---")
    print(Q_est)

    print("-------")
    print(np.abs(Q - Q_est))
    # Check error
    assert Q_est == pytest.approx(Q, rel=0.01)


def test_lsvi_random_exploration():
    env = GridWorld(nrows=2, ncols=2, walls=(), success_probability=0.95)
    env.reseed(123)

    def feature_map_fn(_env):
        return OneHotFeatureMap(_env.observation_space.n, _env.action_space.n)

    agent = LSVIUCBAgent(
        env,
        feature_map_fn=feature_map_fn,
        horizon=20,
        gamma=0.99,
        reg_factor=1e-5,
        bonus_scale_factor=0.0,
    )
    agent.reseed(123)
    agent.fit(budget=250)

    # estimated Q
    S = env.observation_space.n
    Q_est = agent._run_lsvi(bonus_factor=0.0)[0, :].reshape((S, -1))

    # near optimal Q
    agent_opt = ValueIterationAgent(env, gamma=0.99, horizon=20)
    agent_opt.fit()
    Q = agent_opt.Q[0, :, :]

    print(Q)
    print("---")
    print(Q_est)

    print("-------")
    print(np.abs(Q - Q_est))
    # Check error
    assert np.abs(Q - Q_est).mean() < 0.1


def test_lsvi_optimism():
    env = GridWorld(nrows=2, ncols=2, walls=())

    def feature_map_fn(_env):
        return OneHotFeatureMap(_env.observation_space.n, _env.action_space.n)

    agent = LSVIUCBAgent(
        env,
        gamma=0.99,
        feature_map_fn=feature_map_fn,
        horizon=3,
        bonus_scale_factor=3,
        reg_factor=0.000001,
    )
    agent.fit(budget=250)

    # near optimal Q
    agent_opt = ValueIterationAgent(env, gamma=0.99, horizon=3)
    agent_opt.fit()
    Q = agent_opt.Q[0, :, :]

    # optimistic Q
    S = env.observation_space.n
    A = env.action_space.n
    Q_optimistic = np.zeros((S, A))
    for ss in range(S):
        Q_optimistic[ss, :] = agent._compute_q_vec(
            agent.w_vec[0, :], ss, agent.bonus_scale_factor
        )

    print(Q)
    print(Q_optimistic)
    assert (Q_optimistic - Q).min() >= -1e-5



from rlberry.agents import AgentWithSimplePolicy


class RandomAgent(AgentWithSimplePolicy):
    name = "RandomAgent"

    def __init__(self, env, **kwargs):
        AgentWithSimplePolicy.__init__(self, env, **kwargs)

    def fit(self, budget=100, **kwargs):
        observation, info = self.env.reset()
        for ep in range(budget):
            action = self.policy(observation)
            observation, reward, done, _, _ = self.env.step(action)

    def policy(self, observation):
        return self.env.action_space.sample()  # choose an action at random



from abc import ABC, abstractmethod


class FeatureMap(ABC):
    """
    Class representing a feature map, from (observation, action) pairs
    to numpy arrays.

    Attributes
    ----------
    shape : tuple
        Shape of feature array.

    Methods
    --------
    map()
        Maps a (observation, action) pair to a numpy array.
    """

    def __init__(self):
        ABC.__init__(self)
        self.shape = ()

    @abstractmethod
    def map(self, observation, action):
        """
        Maps a (observation, action) pair to a numpy array.
        """
        pass



from setuptools import setup, find_packages
import os

ver_file = os.path.join("rlberry_scool", "_version.py")
with open(ver_file) as f:
    exec(f.read())

packages = find_packages(exclude=["docs"])

#
# Base installation (interface only)
#
install_requires = ["rlberry"]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="rlberry-scool",
    version=__version__,
    description="Teaching Reinforcement Learning made easy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Inria Scool team",
    url="https://github.com/rlberry-py",
    license="MIT",
    packages=packages,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires,
    zip_safe=False,
)



import numpy as np
import gymnasium as gym


def binary_search_nd(x_vec, bins):
    """n-dimensional binary search

    Parameters
    -----------
    x_vec : numpy.ndarray
        numpy 1d array to be searched in the bins
    bins : list
        list of numpy 1d array, bins[d] = bins of the d-th dimension


    Returns
    --------
    index (int) corresponding to the position of x in the partition
    defined by the bins.
    """
    dim = len(bins)
    flat_index = 0
    aux = 1
    assert dim == len(x_vec), "dimension mismatch in binary_search_nd()"
    for dd in range(dim):
        index_dd = np.searchsorted(bins[dd], x_vec[dd], side="right") - 1
        assert np.all(index_dd != -1), "error in binary_search_nd()"
        flat_index += aux * index_dd
        aux *= len(bins[dd]) - 1
    return flat_index


def unravel_index_uniform_bin(flat_index, dim, n_per_dim):
    index = []
    aux_index = flat_index
    for _ in range(dim):
        index.append(aux_index % n_per_dim)
        aux_index = aux_index // n_per_dim
    return tuple(index)


class DiscretizeStateWrapper(gym.ObservationWrapper):
    """
    Discretize an environment with continuous states and discrete actions.
    """

    def __init__(self, env, n_bins):
        # initialize base class
        super().__init__(env)

        self.n_bins = n_bins
        # initialize bins
        assert n_bins > 0, "DiscretizeStateWrapper requires n_bins > 0"
        n_states = 1
        tol = 1e-8
        self.dim = len(self.env.observation_space.low)
        n_states = n_bins ** self.dim
        self._bins = []
        self._open_bins = []
        for dd in range(self.dim):
            range_dd = (
                self.env.observation_space.high[dd] - self.env.observation_space.low[dd]
            )
            epsilon = range_dd / n_bins
            bins_dd = []
            for bb in range(n_bins + 1):
                val = self.env.observation_space.low[dd] + epsilon * bb
                bins_dd.append(val)
            self._open_bins.append(tuple(bins_dd[1:]))
            bins_dd[-1] += tol  # "close" the last interval
            self._bins.append(tuple(bins_dd))

            # set observation space
        self.observation_space = gym.spaces.Discrete(n_states)

        # List of discretized states
        self.discretized_states = np.zeros((self.dim, n_states))
        for ii in range(n_states):
            self.discretized_states[:, ii] = self.get_continuous_state(ii, False)

    def reset(self, **kwargs):
        tmp_state, _ = self.env.reset()
        return self.get_discrete_state(tmp_state), {}

    def step(self, action):
        next_state, reward, term, trunc, info = self.env.step(action)
        next_state = binary_search_nd(next_state, self._bins)
        return next_state, reward, term, trunc, info

    def sample(self, discrete_state, action):
        # map disctete state to continuous one
        assert self.observation_space.contains(discrete_state)
        continuous_state = self.get_continuous_state(discrete_state, randomize=True)
        # sample in the true environment
        next_state, reward, term, trunc, info = self.env.sample(
            continuous_state, action
        )
        # discretize next state
        next_state = binary_search_nd(next_state, self._bins)

        return next_state, reward, term, trunc, info

    def get_discrete_state(self, continuous_state):
        return binary_search_nd(continuous_state, self._bins)

    def get_continuous_state(self, discrete_state, randomize=False):
        assert (
            discrete_state >= 0 and discrete_state < self.observation_space.n
        ), "invalid discrete_state"
        # get multi-index
        index = unravel_index_uniform_bin(discrete_state, self.dim, self.n_bins)

        # get state
        continuous_state = np.zeros(self.dim)
        for dd in range(self.dim):
            continuous_state[dd] = self._bins[dd][index[dd]]
            if randomize:
                range_dd = (
                    self.env.observation_space.high[dd]
                    - self.env.observation_space.low[dd]
                )
                epsilon = range_dd / self.n_bins
                continuous_state[dd] += epsilon * self.rng.uniform()
        return continuous_state



def get_discrete_mountain_car_env():
    env_with_continuous_states = gym.make("MountainCar-v0", render_mode="rgb_array")
    
    env = DiscretizeStateWrapper(env_with_continuous_states, n_bins=10)
    return env



import numpy as np
import pytest

from rlberry.envs import gym_make, PipelineEnv
from rlberry_scool.envs.finite import Chain
from rlberry_scool.envs.finite import GridWorld
from rlberry_scool.envs.finite import get_discrete_mountain_car_env
from rlberry.rendering.render_interface import RenderInterface2D

classes = [
    GridWorld,
    Chain,
]


@pytest.mark.parametrize("ModelClass", classes)
def test_instantiation(ModelClass):
    env = ModelClass()

    if env.is_online():
        for _ in range(2):
            state, info = env.reset()
            for _ in range(50):
                assert env.observation_space.contains(state)
                action = env.action_space.sample()
                next_s, _, _, _, _ = env.step(action)
                state = next_s

    if env.is_generative():
        for _ in range(100):
            state = env.observation_space.sample()
            action = env.action_space.sample()
            next_s, _, _, _, _ = env.sample(state, action)
            assert env.observation_space.contains(next_s)


@pytest.mark.parametrize("ModelClass", classes)
def test_rendering_calls(ModelClass):
    env = ModelClass()
    if isinstance(env, RenderInterface2D):
        _ = env.get_background()
        _ = env.get_scene(env.observation_space.sample())


def test_gridworld_aux_functions():
    env = GridWorld(
        nrows=5, ncols=8, walls=((1, 1),), reward_at={(4, 4): 1, (4, 3): -1}
    )
    env.log()  # from FiniteMDP
    env.render_ascii()  # from GridWorld
    vals = np.arange(env.observation_space.n)
    env.display_values(vals)
    env.print_transition_at(0, 0, "up")

    layout = env.get_layout_array(vals, fill_walls_with=np.inf)
    for rr in range(env.nrows):
        for cc in range(env.ncols):
            if (rr, cc) in env.walls:
                assert layout[rr, cc] == np.inf
            else:
                assert layout[rr, cc] == vals[env.coord2index[(rr, cc)]]


def test_gridworld_from_layout():
    layout = """
    IOOOO # OOOOO  O OOOOR
    OOOOO # OOOOO  # OOOOO
    OOOOO O OOOOO  # OOTOO
    OOOOO # OOOOO  # OOOOO
    IOOOO # OOOOO  # OOOOr"""
    env = GridWorld.from_layout(layout)
    env.reset()


def test_pipeline():
    from rlberry.wrappers import RescaleRewardWrapper
    from rlberry.wrappers.discretize_state import DiscretizeStateWrapper

    env_ctor, env_kwargs = PipelineEnv, {
        "env_ctor": gym_make,
        "env_kwargs": {"id": "Acrobot-v1"},
        "wrappers": [(RescaleRewardWrapper, {"reward_range": (0, 1)})],
    }
    env = env_ctor(**env_kwargs)
    _, reward, _, _, _ = env.step(0)
    assert (reward <= 1) and (reward >= 0)

    env_ctor, env_kwargs = PipelineEnv, {
        "env_ctor": gym_make,
        "env_kwargs": {"id": "Acrobot-v1"},
        "wrappers": [
            (RescaleRewardWrapper, {"reward_range": (0, 1)}),
            (DiscretizeStateWrapper, {"n_bins": 10}),
        ],
    }
    env = env_ctor(**env_kwargs)
    # check that wrapped in the right order
    assert isinstance(
        env.env, RescaleRewardWrapper
    ), "the environments in Pipeline env may not be wrapped in order"
    assert isinstance(env.env.env, DiscretizeStateWrapper)


def test_discrete_mc():
    env = get_discrete_mountain_car_env()
    env.reset()
    done = False
    while not done:
        a = env.action_space.sample()
        _, _, term, trunc, _ = env.step(a)
        done = term or trunc
